{{ COGdesc.spin
================

Part1:
======
Purpose of this is to build a "directory" of PASM drivers that are placed in upper part of
the EEPROM. The 64kByte EEPROM can store up to 15 different drivers. The PASM drivers are
expected at the beginning of the upper 32kB of the EEPROM whereas the COG-descriptor-table
is placed at the end of the EEPROM.

At the very end of the EEPROM the string "COGdesc x.x" is found, where x.x is the version
number of the descriptor table. In this way it can be checked whether the loader understands
the content of this table correctly.

The COG code itself has to be saved in EEPROM using the PASMstoreE_x.spin or the CogOS.
These could already be coded to match your hardware setup. Say you have the SD card at
pins 12-15, you can already 'hardcode' this in the PASMstore. The advantage:
   You don't need hardware specific setup in all applications that use this driver concept.
   This allows to get some code from someone else and run it without changes - except the
   clock setting.


Part 2:
=======
This part is used to store a list of currently loaded drivers. This is for applications
running in a OS or bootloader environment. Here the OS/BL contains the code to load all
drivers needed by the application, whereas the application only needs the code of the
high level driver and a small piece of code which can read the list of drivers.
This saves some additional HUB-RAM.

Part 3:
=======
Code to reserve memory from upper part of the HUB-RAM.
   
}}
con
  NOT_FOUND     = -1
  WRONG_VERSION = -2
  IS_RUNNING    = -3

  FLAG_RUNNING   = $01
  FLAG_SUSPENDED = $02
  FLAG_LINKED    = $04

  EEP_ENTRY_SIZE    = 20
  EEP_TAB_SIZE      = 16 * EEP_ENTRY_SIZE

  RAM_ENTRY_SIZE    = 12
  
  COG_DESC_TAB  = $8000 - 8

  #0, DEV_EEPROM, DEV_FRAM
  
  EEPROM        = $a0
  EEP_BASE      = $8000
  FRAM          = $a2
  FRAM_BASE     = $0000

  DESC_TAB      = $7e00

obj
  i2c:    "Basic_I2C_Driver"

var
  long CogDescr[5]
  word tabAddress
  word buffer_adr
  word par_adr
  byte numOfEntries, numOfDrivers
  byte ramDesc
  byte parMng
  byte device
  word dev_base

pub initRAMDesc( numOfDesc )
  ramDesc := true
  ' in case an application stopped, the descriptor table is still there
  if long[COG_DESC_TAB] & $ffff_ff00 <> ram_id
    long[COG_DESC_TAB] := ram_id
    long[COG_DESC_TAB+4] := COG_DESC_TAB - RAM_ENTRY_SIZE*numOfDesc - 4
      
pri setParAdr
  ' descriptor table initialized as well?
  if ( long[COG_DESC_TAB] & $ffff_ff00 ) == RAM_id 
    par_adr := long[COG_DESC_TAB+4] - 4
  else
    par_adr := COG_DESC_TAB - 8
        
pub getMaxRAM | padr
  ' descriptor table initialized as well?
  'if ( long[COG_DESC_TAB] & $ffff_ff00 ) == RAM_id 
  '  padr := long[COG_DESC_TAB+4] - 4
  'else
  '  padr := COG_DESC_TAB - 8
  setParAdr

  ' in case an application stopped, the parameter table is already there
  if long[ padr ] & $ffff_ff00 == PAR_id & $ffff_ff00
    return long[ padr + 4 ] - 1
  else
    return padr+7 

pub initPARmng
  parMng := true
  ' descriptor table initialized as well?
  'if ( long[COG_DESC_TAB] & $ffff_ff00 ) == RAM_id 
  '  par_adr := long[COG_DESC_TAB+4] - 4
  'else
  '  par_adr := COG_DESC_TAB - 8
  setParAdr

  ' in case an application stopped, the parameter table is already there
  if long[ par_adr ] & $ffff_ff00 <> PAR_id 
    long[ par_adr ] := PAR_id
    long[ par_adr + 4 ] := par_adr

pub parReserve( size )
  long[ par_adr + 4 ] -= size<<2
  byte[ par_adr ]++
  longfill( long[ par_adr + 4 ], 0, size )
  return long[ par_adr + 4 ] 
  
pub initDesc( buf, d ) | i
  ' read the max. possible number of bytes for the descriptor table
  i2c.Initialize(i2c#BootPin)

  case d
    DEV_EEPROM:
      device := EEPROM
      dev_base := EEP_BASE
    DEV_FRAM:
      device := FRAM
      dev_base := FRAM_BASE

  i2c.ReadPage(i2c#BootPin, device, dev_base+DESC_TAB, buf, 512)

  buffer_adr := buf
  tabAddress := buf + $00b4
  numOfEntries := byte[ tabAddress + EEP_TAB_SIZE + 11 ]

  ' check that the loaded EEPROM data really contains the table
  repeat i from 0 to 10
    if byte[ @eep_id + i ] <> byte[ tabAddress + EEP_TAB_SIZE + i ]
      abort -1

  return numOfEntries

pub getDevice
  if device==EEPROM
    return DEV_EEPROM
  if device==FRAM
    return DEV_FRAM
  return 0
    
pub clear
  bytemove( buffer_adr+500, @EEP_id, 12 )

pub startDriver(pasm, parameter) : retVal
  ' check whether the actual driver already runs or not
  if long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 4] & FLAG_RUNNING
    ' if so, create a new table entry
    byte[ COG_DESC_TAB ] ++
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*byte[ COG_DESC_TAB ] ] := long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc ]
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*byte[ COG_DESC_TAB ] + 4 ] := 0
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*byte[ COG_DESC_TAB ] + 8 ] := long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 8 ]
    ramDesc := byte[ COG_DESC_TAB ]

    ' start the PASM code
  retVal := cognew( pasm, parameter )

  ' and set the running-flag/COG ID/address of parameter block
  if retVal>-1
    retVal++
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 4] |= FLAG_RUNNING
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 4] |= retVal << 8
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 4] |= parameter << 16
    
pub addEntry( name, version, cognr, parcnt, paradr )
    byte[ COG_DESC_TAB ] ++
    ramDesc := byte[ COG_DESC_TAB ]
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc ] := Hash( name, version )
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 4] |= FLAG_RUNNING
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 4] |= cognr << 8
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 4] |= paradr << 16
    long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 8 ] := parcnt
    
pub saveTable | cpylen, frombuf, tobuf, start_time, i
  i2c.Initialize(i2c#BootPin)
  cpylen := 512
  frombuf := buffer_adr
  tobuf := dev_base + DESC_TAB
  repeat while cpylen>0
    i := cpylen<#64
    if !i2c.WritePage(i2c#BootPin, device, tobuf, frombuf, i)
      start_time := cnt ' prepare to check for a timeout
      repeat while i2c.WriteWait(i2c#BootPin, device, tobuf)
        if cnt - start_time > clkfreq / 10
          quit
    else
      quit
    cpylen-=64
    frombuf+=64
    tobuf+=64

pub saveDriver( slot, pasmAdr ) | cpylen, frombuf, tobuf, start_time, i
  if slot<0 or slot=>EEP_ENTRY_SIZE
    return
    
  i2c.Initialize(i2c#BootPin)
  cpylen := 2048
  frombuf := pasmAdr
  tobuf := dev_base + slot << 11
  repeat while cpylen>0
    i := cpylen<#64
    if !i2c.WritePage(i2c#BootPin, device, tobuf, frombuf, i)
      start_time := cnt ' prepare to check for a timeout
      repeat while i2c.WriteWait(i2c#BootPin, device, tobuf)
        if cnt - start_time > clkfreq / 10
          quit
    else
      quit
    cpylen-=64
    frombuf+=64
    tobuf+=64

pub loadDriver( name, version ) : retVal | found, i
  ' initialize the return value
  retVal := NOT_FOUND

  ' if RAM descriptor table is switched on
  if ramDesc
    ' get the number of drivers already loaded
    numOfDrivers := byte[ COG_DESC_TAB ]
    ' convert the driver name to a hash value for saving memory
    found := Hash( name, version )
    ' driver already loaded? Then return
    repeat i from 1 to numOfDrivers
      if found == long[ COG_DESC_TAB - RAM_ENTRY_SIZE*i ]
        return IS_RUNNING

  ' not already loaded, so search for the driver in the EEPROM descriptor table      
  found:=0
  if numOfEntries>0
    repeat i from 0 to numOfEntries-1
      if strCmp( name, tabAddress + EEP_TAB_SIZE-12 - EEP_ENTRY_SIZE*i, 11 )
        retVal := WRONG_VERSION
      ' here we found the string and have to check the version
        if version == word[ tabAddress + EEP_TAB_SIZE-18 - EEP_ENTRY_SIZE*i ]
          found:=-1
          quit

  ' if the driver has been found, load it from the according slot
  if found
    retVal := i
    ' get the descriptor, so that the other getters work
    getDescrFrom( i )
    i2c.ReadPage(i2c#BootPin, device, dev_base + i<<11 , buffer_adr + 512, 2048)
    ' if RAM descriptor table is switched on, set the descriptor
    if ramDesc
      found := Hash( name, version )
      byte[ COG_DESC_TAB ] ++
      ramDesc := byte[ COG_DESC_TAB ]
      long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc ] := found
      long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 4] := 0
      long[ COG_DESC_TAB - RAM_ENTRY_SIZE*ramDesc + 8] := word[ tabAddress + EEP_TAB_SIZE- EEP_ENTRY_SIZE - EEP_ENTRY_SIZE*i ]
      

pub Hash(string_ptr, version) : Result | x
 result := 0
 repeat strsize(string_ptr)
   Result := (Result << 4) + byte[string_ptr++]
   x := Result & $F0_00_00_00
   if (x <> 0)
    Result := Result ^ (x >> 24)
   Result := (result & !x ) ^ version


pub loadSlot( slot )
    i2c.ReadPage(i2c#BootPin, device, dev_base + slot<<11 , buffer_adr + 512, 2048)

pub strCmp( adrA, adrB, maxLen ) | i
  repeat i from 0 to maxLen
    if byte[ adrA + i ] <> byte[ adrB + i ]
      return 0
    if byte[ adrA + i ] == 0 and byte[ adrB + i ] == 0
      return -1
  return -1

pub clearLastSlot | dAdr
  dAdr := tabAddress  + EEP_TAB_SIZE - EEP_ENTRY_SIZE*( numOfEntries )
  longfill( dAdr, 0, 5 )
  numOfEntries--
  byte[ tabAddress + EEP_TAB_SIZE + 11 ] := numOfEntries
  
pub putDescription : retVal | found, i
  found:=0
  if numOfEntries>0
    repeat i from 0 to numOfEntries-1
      if strCmp( @CogDescr+8, tabAddress + EEP_TAB_SIZE - 12 - EEP_ENTRY_SIZE*i, 11 )
      ' here we found the string and have to check the version
        if word[ @CogDescr + 2 ] == word[ tabAddress + EEP_TAB_SIZE - 18 - EEP_ENTRY_SIZE*i ]
          found:=-1
          quit

  if !found
    if numOfEntries<EEP_ENTRY_SIZE
      putDescrTo( numOfEntries )
      retVal := numOfEntries
      numOfEntries++
      byte[ tabAddress + EEP_TAB_SIZE+11 ] := numOfEntries
    else
      retVal := -1
  else
    putDescrTo( i )
    retVal := i
  
pub setName( strPtr ) | ctr1, ctr2
  ctr1 := 0
  ctr2 := 0  
  repeat while ctr1<12
    byte[ @CogDescr + 8 + ctr1 ]:=byte[ strPtr + ctr2 ]
    ctr1++
    if byte[ strPtr + ctr2 ] <> 0 
      ctr2++
      
pub setVAdr( vAdr )
  long[@CogDescr+4] := vAdr
                                             
pub setVersion( version )
  word[ @CogDescr+2 ] := version & $ffff
  
pub setPSize( size )
  word[ @CogDescr ] := size & $ffff
  
pub putDescrTo( slot ) | dAdr
  dAdr := tabAddress  + EEP_TAB_SIZE - EEP_ENTRY_SIZE*(slot+1)
  longmove( dAdr, @CogDescr, 5 )
  
pub getDescrFrom( slot ) | dAdr
  dAdr := tabAddress + EEP_TAB_SIZE - EEP_ENTRY_SIZE*(slot+1)
  longmove( @CogDescr, dAdr, 5 )
  
pub getVAdr
  return long[@CogDescr+4]

pub getPSize
  return word[ @CogDescr ]

pub getName( strPtr ) | ctr1, ctr2  
  ctr1 := 0
  ctr2 := 0  
  repeat while ctr2<12
    byte[ strPtr + ctr2 ]:=byte[ @CogDescr + 8 + ctr1 ]
    ctr2++
    if byte[ @CogDescr + 8 + ctr1 ] <> 0 
      ctr1++

pub getVersion
  return word[@CogDescr+2] & $ffff

pub getDrvCnt
  return byte[ tabAddress + EEP_TAB_SIZE+11 ]

pub findRAMDrv(startFrom, hashVal ) : i
  repeat i from startFrom to getRAMDrvCnt
    if getRAMHash( i ) == hashVal
      if (getRAMStatus( i ) & FLAG_LINKED) == 0
        return
  return -1

pub setRAMStatus( i, status )
  long[ COG_DESC_TAB - RAM_ENTRY_SIZE*i + 4 ] := (long[ COG_DESC_TAB - RAM_ENTRY_SIZE*i + 4 ] & $ffff_ff00) | (status & $ff)

pub getRAMDrvCnt
  return byte[ COG_DESC_TAB ]

pub getRAMStatus( i )
  return long[ COG_DESC_TAB - RAM_ENTRY_SIZE*i + 4] & $FF
 
pub getRAMParAddress( i )
  return long[ COG_DESC_TAB - RAM_ENTRY_SIZE*i + 4] >> 16 & $ffff

pub getRAMParSize( i )
  return long[ COG_DESC_TAB - RAM_ENTRY_SIZE*i + 8] & $ffff

pub getRAMBufferSize( i )

pub getRAMHash( i )
  return long[ COG_DESC_TAB - RAM_ENTRY_SIZE*i ]

pub getRAMCOG( i )
  return long[ COG_DESC_TAB - RAM_ENTRY_SIZE*i + 4]>>8 & $ff

pub getID( cAdr ) | ctr
  repeat ctr from 0 to 10
    byte[ cAdr++ ] := byte[ tabAddress + EEP_TAB_SIZE + ctr ] 
  byte[ cAdr ] := 0
  
dat
RAM_id  long  $47_4F_43_00
PAR_id  long  $52_41_50_00
EEP_id  byte  "COGdesc 0.1",0 